@@ -19,6 +19,7 @@
 #include <arpa/inet.h>
 #include <errno.h>
 #include <malloc.h>
+#include <stdlib.h>
 #include <string.h>
 #include <time.h>
 
@@ -27,9 +28,6 @@
 #include "print.h"
 #include "tlv.h"
 
-#define VERSION_MASK 0x0f
-#define VERSION      0x02
-
 int assume_two_step = 0;
 
 /*
@@ -39,8 +37,8 @@ int assume_two_step = 0;
 
 struct message_storage {
 	unsigned char reserved[MSG_HEADROOM];
-	struct ptp_message msg;
-} PACKED;
+	struct ptp_message msg __attribute__((aligned (8)));
+};
 
 static TAILQ_HEAD(msg_pool, ptp_message) msg_pool = TAILQ_HEAD_INITIALIZER(msg_pool);
 
@@ -80,7 +78,7 @@ static void announce_post_recv(struct announce_msg *m)
 
 static int hdr_post_recv(struct ptp_header *m)
 {
-	if ((m->ver & VERSION_MASK) != VERSION)
+	if ((m->ver & MAJOR_VERSION_MASK) != PTP_MAJOR_VERSION)
 		return -EPROTO;
 	m->messageLength = ntohs(m->messageLength);
 	m->correction = net2host64(m->correction);
@@ -102,13 +100,13 @@ static uint8_t *msg_suffix(struct ptp_message *m)
 {
 	switch (msg_type(m)) {
 	case SYNC:
-		return NULL;
+		return m->sync.suffix;
 	case DELAY_REQ:
 		return m->delay_req.suffix;
 	case PDELAY_REQ:
-		return NULL;
+		return m->pdelay_req.suffix;
 	case PDELAY_RESP:
-		return NULL;
+		return m->pdelay_resp.suffix;
 	case FOLLOW_UP:
 		return m->follow_up.suffix;
 	case DELAY_RESP:
