@@ -39,6 +39,7 @@
 
 #include "clockadj.h"
 #include "clockcheck.h"
+#include "contain.h"
 #include "ds.h"
 #include "fsm.h"
 #include "missing.h"
@@ -46,7 +47,7 @@
 #include "ntpshm.h"
 #include "phc.h"
 #include "pi.h"
-#include "pmc_common.h"
+#include "pmc_agent.h"
 #include "print.h"
 #include "servo.h"
 #include "sk.h"
@@ -62,12 +63,8 @@
 #define NS_PER_SEC 1000000000LL
 
 #define PHC_PPS_OFFSET_LIMIT 10000000
-#define PMC_UPDATE_INTERVAL (60 * NS_PER_SEC)
-#define PMC_SUBSCRIBE_DURATION 180	/* 3 minutes */
-/* Note that PMC_SUBSCRIBE_DURATION has to be longer than
- * PMC_UPDATE_INTERVAL otherwise subscription will time out before it is
- * renewed.
- */
+
+#define MAX_DST_CLOCKS 128
 
 struct clock {
 	LIST_ENTRY(clock) list;
@@ -105,17 +102,11 @@ struct phc2sys_private {
 	enum servo_type servo_type;
 	int phc_readings;
 	double phc_interval;
-	int sync_offset;
 	int forced_sync_offset;
-	int utc_offset_traceable;
-	int leap;
 	int kernel_leap;
-	struct pmc *pmc;
-	int pmc_ds_requested;
-	uint64_t pmc_last_update;
 	int state_changed;
-	int clock_identity_set;
-	struct ClockIdentity clock_identity;
+	int free_running;
+	struct pmc_agent *agent;
 	LIST_HEAD(port_head, port) ports;
 	LIST_HEAD(clock_head, clock) clocks;
 	LIST_HEAD(dst_clock_head, clock) dst_clocks;
@@ -124,18 +115,9 @@ struct phc2sys_private {
 
 static struct config *phc2sys_config;
 
-static int update_pmc(struct phc2sys_private *priv, int subscribe);
 static int clock_handle_leap(struct phc2sys_private *priv,
 			     struct clock *clock,
 			     int64_t offset, uint64_t ts);
-static int run_pmc_get_utc_offset(struct phc2sys_private *priv,
-				  int timeout);
-static void run_pmc_events(struct phc2sys_private *priv);
-
-static int normalize_state(int state);
-static int run_pmc_port_properties(struct phc2sys_private *priv,
-				   int timeout, unsigned int port,
-				   int *state, int *tstamping, char *iface);
 
 static struct servo *servo_add(struct phc2sys_private *priv,
 			       struct clock *clock)
@@ -146,9 +128,6 @@ static struct servo *servo_add(struct phc2sys_private *priv,
 
 	clockadj_init(clock->clkid);
 	ppb = clockadj_get_freq(clock->clkid);
-	/* The reading may silently fail and return 0, reset the frequency to
-	   make sure ppb is the actual frequency of the clock. */
-	clockadj_set_freq(clock->clkid, ppb);
 	if (clock->clkid == CLOCK_REALTIME) {
 		sysclk_set_leap(0);
 		max_ppb = sysclk_max_freq();
@@ -172,14 +151,21 @@ static struct servo *servo_add(struct phc2sys_private *priv,
 	return servo;
 }
 
-static struct clock *clock_add(struct phc2sys_private *priv, char *device)
+static struct clock *clock_add(struct phc2sys_private *priv, const char *device,
+			       int phc_index)
 {
 	struct clock *c;
 	clockid_t clkid = CLOCK_INVALID;
-	int phc_index = -1;
+	char phc_device[19];
 
 	if (device) {
-		clkid = posix_clock_open(device, &phc_index);
+		if (phc_index >= 0) {
+			snprintf(phc_device, sizeof(phc_device), "/dev/ptp%d",
+				 phc_index);
+			clkid = posix_clock_open(phc_device, &phc_index);
+		} else {
+			clkid = posix_clock_open(device, &phc_index);
+		}
 		if (clkid == CLOCK_INVALID)
 			return NULL;
 	}
@@ -220,9 +206,6 @@ static struct clock *clock_add(struct phc2sys_private *priv, char *device)
 		}
 	}
 
-	if (clkid != CLOCK_INVALID)
-		c->servo = servo_add(priv, c);
-
 	if (clkid != CLOCK_INVALID && clkid != CLOCK_REALTIME)
 		c->sysoff_method = sysoff_probe(CLOCKID_TO_FD(clkid),
 						priv->phc_readings);
@@ -279,7 +262,7 @@ static struct port *port_get(struct phc2sys_private *priv, unsigned int number)
 }
 
 static struct port *port_add(struct phc2sys_private *priv, unsigned int number,
-			     char *device)
+			     char *device, int phc_index)
 {
 	struct port *p;
 	struct clock *c = NULL, *tmp;
@@ -296,7 +279,7 @@ static struct port *port_add(struct phc2sys_private *priv, unsigned int number,
 		}
 	}
 	if (!c) {
-		c = clock_add(priv, device);
+		c = clock_add(priv, device, phc_index);
 		if (!c)
 			return NULL;
 	}
@@ -314,34 +297,37 @@ static struct port *port_add(struct phc2sys_private *priv, unsigned int number,
 static void clock_reinit(struct phc2sys_private *priv, struct clock *clock,
 			 int new_state)
 {
-	int phc_index = -1, phc_switched = 0;
-	int state, timestamping, ret = -1;
+	int err = -1, phc_index = -1, phc_switched = 0, timestamping;
+	char iface[IFNAMSIZ], phc_device[19];
+	enum port_state state;
 	struct port *p;
-	struct servo *servo;
-	struct sk_ts_info ts_info;
-	char iface[IFNAMSIZ];
 	clockid_t clkid = CLOCK_INVALID;
 
 	LIST_FOREACH(p, &priv->ports, list) {
-		if (p->clock == clock) {
-			ret = run_pmc_port_properties(priv, 1000, p->number,
-					              &state, &timestamping,
-						      iface);
-			if (ret > 0)
-				p->state = normalize_state(state);
+		if (p->clock != clock) {
+			continue;
 		}
+		err = pmc_agent_query_port_properties(priv->agent, 1000,
+						      p->number, &state,
+						      &timestamping, &phc_index,
+						      iface);
+		if (!err) {
+			p->state = port_state_normalize(state);
+		}
+		break;
 	}
 
-	if (ret > 0 && timestamping != TS_SOFTWARE) {
+	if (!err && timestamping != TS_SOFTWARE) {
 		/* Check if device changed */
 		if (strcmp(clock->device, iface)) {
 			free(clock->device);
 			clock->device = strdup(iface);
 		}
 		/* Check if phc index changed */
-		if (!sk_get_ts_info(clock->device, &ts_info) &&
-		    clock->phc_index != ts_info.phc_index) {
-			clkid = posix_clock_open(clock->device, &phc_index);
+		if (clock->phc_index != phc_index) {
+			snprintf(phc_device, sizeof(phc_device), "/dev/ptp%d",
+				 phc_index);
+			clkid = posix_clock_open(phc_device, &phc_index);
 			if (clkid == CLOCK_INVALID)
 				return;
 
@@ -349,10 +335,9 @@ static void clock_reinit(struct phc2sys_private *priv, struct clock *clock,
 			clock->clkid = clkid;
 			clock->phc_index = phc_index;
 
-			servo = servo_add(priv, clock);
-			if (servo) {
+			if (clock->servo) {
 				servo_destroy(clock->servo);
-				clock->servo = servo;
+				clock->servo = NULL;
 			}
 
 			phc_switched = 1;
@@ -360,7 +345,8 @@ static void clock_reinit(struct phc2sys_private *priv, struct clock *clock,
 	}
 
 	if (new_state == PS_MASTER || phc_switched) {
-		servo_reset(clock->servo);
+		if (clock->servo)
+			servo_reset(clock->servo);
 		clock->servo_state = SERVO_UNLOCKED;
 
 		if (clock->offset_stats) {
@@ -369,6 +355,9 @@ static void clock_reinit(struct phc2sys_private *priv, struct clock *clock,
 			stats_reset(clock->delay_stats);
 		}
 	}
+
+	pr_debug("%s: state change %s -> %s", clock->device,
+		 ps_str[clock->state], ps_str[new_state]);
 }
 
 static struct clock *find_dst_clock(struct phc2sys_private *priv,
@@ -437,19 +426,6 @@ static void reconfigure(struct phc2sys_private *priv)
 		}
 		last = c;
 	}
-	if (dst_cnt > 1 && !src) {
-		if (!rt || rt->dest_only) {
-			priv->master = last;
-			/* Reset to original state in next reconfiguration. */
-			priv->master->new_state = priv->master->state;
-			priv->master->state = PS_SLAVE;
-			if (rt)
-				rt->state = PS_SLAVE;
-			pr_info("no source, selecting %s as the default clock",
-				last->device);
-			return;
-		}
-	}
 	if (src_cnt > 1) {
 		pr_info("multiple master clocks available, postponing sync...");
 		priv->master = NULL;
@@ -465,6 +441,21 @@ static void reconfigure(struct phc2sys_private *priv)
 		priv->master = NULL;
 		return;
 	}
+
+	if (dst_cnt > 1 && !src) {
+		if (!rt || rt->dest_only) {
+			priv->master = last;
+			/* Reset to original state in next reconfiguration. */
+			priv->master->new_state = priv->master->state;
+			priv->master->state = PS_SLAVE;
+			if (rt)
+				rt->state = PS_SLAVE;
+			pr_info("no source, selecting %s as the default clock",
+				last->device);
+			return;
+		}
+	}
+
 	if ((!src_cnt && (!rt || rt->dest_only)) ||
 	    (!dst_cnt && !rt)) {
 		pr_info("nothing to synchronize");
@@ -486,37 +477,6 @@ static void reconfigure(struct phc2sys_private *priv)
 	pr_info("selecting %s as the master clock", src->device);
 }
 
-static int read_phc(clockid_t clkid, clockid_t sysclk, int readings,
-		    int64_t *offset, uint64_t *ts, int64_t *delay)
-{
-	struct timespec tdst1, tdst2, tsrc;
-	int i;
-	int64_t interval, best_interval = INT64_MAX;
-
-	/* Pick the quickest clkid reading. */
-	for (i = 0; i < readings; i++) {
-		if (clock_gettime(sysclk, &tdst1) ||
-				clock_gettime(clkid, &tsrc) ||
-				clock_gettime(sysclk, &tdst2)) {
-			pr_err("failed to read clock: %m");
-			return 0;
-		}
-
-		interval = (tdst2.tv_sec - tdst1.tv_sec) * NS_PER_SEC +
-			tdst2.tv_nsec - tdst1.tv_nsec;
-
-		if (best_interval > interval) {
-			best_interval = interval;
-			*offset = (tdst1.tv_sec - tsrc.tv_sec) * NS_PER_SEC +
-				tdst1.tv_nsec - tsrc.tv_nsec + interval / 2;
-			*ts = tdst2.tv_sec * NS_PER_SEC + tdst2.tv_nsec;
-		}
-	}
-	*delay = best_interval;
-
-	return 1;
-}
-
 static int64_t get_sync_offset(struct phc2sys_private *priv, struct clock *dst)
 {
 	int direction = priv->forced_sync_offset;
@@ -568,14 +528,23 @@ static void update_clock_stats(struct clock *clock, unsigned int max_count,
 static void update_clock(struct phc2sys_private *priv, struct clock *clock,
 			 int64_t offset, uint64_t ts, int64_t delay)
 {
-	enum servo_state state;
-	double ppb;
+	enum servo_state state = SERVO_UNLOCKED;
+	double ppb = 0.0;
+
+	if (!clock->servo) {
+		clock->servo = servo_add(priv, clock);
+		if (!clock->servo)
+			return;
+	}
 
 	if (clock_handle_leap(priv, clock, offset, ts))
 		return;
 
 	offset += get_sync_offset(priv, clock);
 
+	if (priv->free_running)
+		goto report;
+
 	if (clock->sanity_check && clockcheck_sample(clock->sanity_check, ts))
 		servo_reset(clock->servo);
 
@@ -586,13 +555,20 @@ static void update_clock(struct phc2sys_private *priv, struct clock *clock,
 	case SERVO_UNLOCKED:
 		break;
 	case SERVO_JUMP:
-		clockadj_step(clock->clkid, -offset);
+		if (clockadj_step(clock->clkid, -offset)) {
+			goto servo_unlock;
+		}
 		if (clock->sanity_check)
 			clockcheck_step(clock->sanity_check, -offset);
 		/* Fall through. */
 	case SERVO_LOCKED:
 	case SERVO_LOCKED_STABLE:
-		clockadj_set_freq(clock->clkid, -ppb);
+		if (clock->sanity_check)
+			clockcheck_freq(clock->sanity_check,
+					clockadj_get_freq(clock->clkid));
+		if (clockadj_set_freq(clock->clkid, -ppb)) {
+			goto servo_unlock;
+		}
 		if (clock->clkid == CLOCK_REALTIME)
 			sysclk_set_sync();
 		if (clock->sanity_check)
@@ -600,6 +576,7 @@ static void update_clock(struct phc2sys_private *priv, struct clock *clock,
 		break;
 	}
 
+report:
 	if (clock->offset_stats) {
 		update_clock_stats(clock, priv->stats_max_count, offset, ppb, delay);
 	} else {
@@ -614,6 +591,11 @@ static void update_clock(struct phc2sys_private *priv, struct clock *clock,
 				offset, state, ppb);
 		}
 	}
+	return;
+
+servo_unlock:
+	servo_reset(clock->servo);
+	clock->servo_state = SERVO_UNLOCKED;
 }
 
 static void enable_pps_output(clockid_t src)
@@ -654,12 +636,13 @@ static int do_pps_loop(struct phc2sys_private *priv, struct clock *clock,
 	int64_t pps_offset, phc_offset, phc_delay;
 	uint64_t pps_ts, phc_ts;
 	clockid_t src = priv->master->clkid;
+	int err;
 
 	priv->master->source_label = "pps";
 
 	if (src == CLOCK_INVALID) {
 		/* The sync offset can't be applied with PPS alone. */
-		priv->sync_offset = 0;
+		pmc_agent_set_sync_offset(priv->agent, 0);
 	} else {
 		enable_pps_output(priv->master->clkid);
 	}
@@ -672,8 +655,13 @@ static int do_pps_loop(struct phc2sys_private *priv, struct clock *clock,
 		/* If a PHC is available, use it to get the whole number
 		   of seconds in the offset and PPS for the rest. */
 		if (src != CLOCK_INVALID) {
-			if (!read_phc(src, clock->clkid, priv->phc_readings,
-				      &phc_offset, &phc_ts, &phc_delay))
+			err = clockadj_compare(src, clock->clkid,
+					       priv->phc_readings,
+					       &phc_offset, &phc_ts,
+					       &phc_delay);
+			if (err == -EBUSY)
+				continue;
+			if (err)
 				return -1;
 
 			/* Convert the time stamp to the PHC time. */
@@ -690,7 +678,7 @@ static int do_pps_loop(struct phc2sys_private *priv, struct clock *clock,
 			pps_offset = pps_ts - phc_ts;
 		}
 
-		if (update_pmc(priv, 0) < 0)
+		if (pmc_agent_update(priv->agent) < 0)
 			continue;
 		update_clock(priv, clock, pps_offset, pps_ts, -1);
 	}
@@ -715,32 +703,31 @@ static int update_needed(struct clock *c)
 	return 0;
 }
 
-static int do_loop(struct phc2sys_private *priv, int subscriptions)
+static int do_loop(struct phc2sys_private *priv)
 {
 	struct timespec interval;
 	struct clock *clock;
 	uint64_t ts;
 	int64_t offset, delay;
+	int err;
 
 	interval.tv_sec = priv->phc_interval;
 	interval.tv_nsec = (priv->phc_interval - interval.tv_sec) * 1e9;
 
 	while (is_running()) {
 		clock_nanosleep(CLOCK_MONOTONIC, 0, &interval, NULL);
-		if (update_pmc(priv, subscriptions) < 0)
-			continue;
 
-		if (subscriptions) {
-			run_pmc_events(priv);
-			if (priv->state_changed) {
-				/* force getting offset, as it may have
-				 * changed after the port state change */
-				if (run_pmc_get_utc_offset(priv, 1000) <= 0) {
-					pr_err("failed to get UTC offset");
-					continue;
-				}
-				reconfigure(priv);
+		if (pmc_agent_update(priv->agent) < 0) {
+			continue;
+		}
+		if (priv->state_changed) {
+			/* force getting offset, as it may have
+			 * changed after the port state change */
+			if (pmc_agent_query_utc_offset(priv->agent, 1000)) {
+				pr_err("failed to get UTC offset");
+				continue;
 			}
+			reconfigure(priv);
 		}
 		if (!priv->master)
 			continue;
@@ -756,99 +743,41 @@ static int do_loop(struct phc2sys_private *priv, int subscriptions)
 			    !strcmp(clock->device, priv->master->device))
 				continue;
 
-			if (!clock->servo) {
-				pr_err("cannot update clock without servo");
-				return -1;
-			}
-
 			if (clock->clkid == CLOCK_REALTIME &&
 			    priv->master->sysoff_method >= 0) {
 				/* use sysoff */
-				if (sysoff_measure(CLOCKID_TO_FD(priv->master->clkid),
-						   priv->master->sysoff_method,
-						   priv->phc_readings,
-						   &offset, &ts, &delay) < 0)
-					return -1;
+				err = sysoff_measure(CLOCKID_TO_FD(priv->master->clkid),
+						     priv->master->sysoff_method,
+						     priv->phc_readings,
+						     &offset, &ts, &delay);
 			} else if (priv->master->clkid == CLOCK_REALTIME &&
 				   clock->sysoff_method >= 0) {
 				/* use reversed sysoff */
-				if (sysoff_measure(CLOCKID_TO_FD(clock->clkid),
-						   clock->sysoff_method,
-						   priv->phc_readings,
-						   &offset, &ts, &delay) < 0)
-					return -1;
-				offset = -offset;
-				ts += offset;
+				err = sysoff_measure(CLOCKID_TO_FD(clock->clkid),
+						     clock->sysoff_method,
+						     priv->phc_readings,
+						     &offset, &ts, &delay);
+				if (!err) {
+					offset = -offset;
+					ts += offset;
+				}
 			} else {
 				/* use phc */
-				if (!read_phc(priv->master->clkid, clock->clkid,
-					      priv->phc_readings,
-					      &offset, &ts, &delay))
-					continue;
+				err = clockadj_compare(priv->master->clkid,
+						       clock->clkid,
+						       priv->phc_readings,
+						       &offset, &ts, &delay);
 			}
+			if (err == -EBUSY)
+				continue;
+			if (err)
+				return -1;
 			update_clock(priv, clock, offset, ts, delay);
 		}
 	}
 	return 0;
 }
 
-static int check_clock_identity(struct phc2sys_private *priv,
-				struct ptp_message *msg)
-{
-	if (!priv->clock_identity_set)
-		return 1;
-	return cid_eq(&priv->clock_identity,
-		       &msg->header.sourcePortIdentity.clockIdentity);
-}
-
-static int is_msg_mgt(struct ptp_message *msg)
-{
-	struct TLV *tlv;
-
-	if (msg_type(msg) != MANAGEMENT)
-		return 0;
-	if (management_action(msg) != RESPONSE)
-		return 0;
-	if (msg_tlv_count(msg) != 1)
-		return 0;
-	tlv = (struct TLV *) msg->management.suffix;
-	if (tlv->type == TLV_MANAGEMENT)
-		return 1;
-	if (tlv->type == TLV_MANAGEMENT_ERROR_STATUS)
-		return -1;
-	return 0;
-}
-
-static int get_mgt_id(struct ptp_message *msg)
-{
-	struct management_tlv *mgt = (struct management_tlv *) msg->management.suffix;
-	return mgt->id;
-}
-
-static void *get_mgt_data(struct ptp_message *msg)
-{
-	struct management_tlv *mgt = (struct management_tlv *) msg->management.suffix;
-	return mgt->data;
-}
-
-static int get_mgt_err_id(struct ptp_message *msg)
-{
-	struct management_error_status *mgt;
-
-	mgt = (struct management_error_status *)msg->management.suffix;
-	return mgt->id;
-}
-
-static int normalize_state(int state)
-{
-	if (state != PS_MASTER && state != PS_SLAVE &&
-	    state != PS_PRE_MASTER && state != PS_UNCALIBRATED) {
-		/* treat any other state as "not a master nor a slave" */
-		state = PS_DISABLED;
-	}
-	return state;
-}
-
 static int clock_compute_state(struct phc2sys_private *priv,
 			       struct clock *clock)
 {
@@ -868,27 +797,28 @@ static int clock_compute_state(struct phc2sys_private *priv,
 	return state;
 }
 
-static int recv_subscribed(struct phc2sys_private *priv,
-			   struct ptp_message *msg, int excluded)
+static int phc2sys_recv_subscribed(void *context, struct ptp_message *msg,
+				   int excluded)
 {
+	struct phc2sys_private *priv = (struct phc2sys_private *) context;
 	int mgt_id, state;
 	struct portDS *pds;
 	struct port *port;
 	struct clock *clock;
 
-	mgt_id = get_mgt_id(msg);
+	mgt_id = management_tlv_id(msg);
 	if (mgt_id == excluded)
 		return 0;
 	switch (mgt_id) {
-	case TLV_PORT_DATA_SET:
-		pds = get_mgt_data(msg);
+	case MID_PORT_DATA_SET:
+		pds = management_tlv_data(msg);
 		port = port_get(priv, pds->portIdentity.portNumber);
 		if (!port) {
 			pr_info("received data for unknown port %s",
 				pid2str(&pds->portIdentity));
 			return 1;
 		}
-		state = normalize_state(pds->portState);
+		state = port_state_normalize(pds->portState);
 		if (port->state != state) {
 			pr_info("port %s changed state",
 				pid2str(&pds->portIdentity));
@@ -905,300 +835,51 @@ static int recv_subscribed(struct phc2sys_private *priv,
 	return 0;
 }
 
-static void send_subscription(struct phc2sys_private *priv)
-{
-	struct subscribe_events_np sen;
-
-	memset(&sen, 0, sizeof(sen));
-	sen.duration = PMC_SUBSCRIBE_DURATION;
-	sen.bitmask[0] = 1 << NOTIFY_PORT_STATE;
-	pmc_send_set_action(priv->pmc, TLV_SUBSCRIBE_EVENTS_NP, &sen, sizeof(sen));
-}
-
-static int init_pmc(struct config *cfg, struct phc2sys_private *priv)
-{
-	char uds_local[MAX_IFNAME_SIZE + 1];
-
-	snprintf(uds_local, sizeof(uds_local), "/var/run/phc2sys.%d",
-		 getpid());
-	priv->pmc = pmc_create(cfg, TRANS_UDS, uds_local, 0,
-			       config_get_int(cfg, NULL, "domainNumber"),
-			       config_get_int(cfg, NULL, "transportSpecific") << 4, 1);
-	if (!priv->pmc) {
-		pr_err("failed to create pmc");
-		return -1;
-	}
-
-	return 0;
-}
-
-/* Return values:
- * 1: success
- * 0: timeout
- * -1: error reported by the other side
- * -2: local error, fatal
- */
-static int run_pmc(struct phc2sys_private *priv, int timeout, int ds_id,
-		   struct ptp_message **msg)
-{
-#define N_FD 1
-	struct pollfd pollfd[N_FD];
-	int cnt, res;
-
-	while (1) {
-		pollfd[0].fd = pmc_get_transport_fd(priv->pmc);
-		pollfd[0].events = POLLIN|POLLPRI;
-		if (!priv->pmc_ds_requested && ds_id >= 0)
-			pollfd[0].events |= POLLOUT;
-
-		cnt = poll(pollfd, N_FD, timeout);
-		if (cnt < 0) {
-			pr_err("poll failed");
-			return -2;
-		}
-		if (!cnt) {
-			/* Request the data set again in the next run. */
-			priv->pmc_ds_requested = 0;
-			return 0;
-		}
-
-		/* Send a new request if there are no pending messages. */
-		if ((pollfd[0].revents & POLLOUT) &&
-		    !(pollfd[0].revents & (POLLIN|POLLPRI))) {
-			switch (ds_id) {
-			case TLV_SUBSCRIBE_EVENTS_NP:
-				send_subscription(priv);
-				break;
-			default:
-				pmc_send_get_action(priv->pmc, ds_id);
-				break;
-			}
-			priv->pmc_ds_requested = 1;
-		}
-
-		if (!(pollfd[0].revents & (POLLIN|POLLPRI)))
-			continue;
-
-		*msg = pmc_recv(priv->pmc);
-
-		if (!*msg)
-			continue;
-
-		if (!check_clock_identity(priv, *msg)) {
-			msg_put(*msg);
-			*msg = NULL;
-			continue;
-		}
-
-		res = is_msg_mgt(*msg);
-		if (res < 0 && get_mgt_err_id(*msg) == ds_id) {
-			priv->pmc_ds_requested = 0;
-			return -1;
-		}
-		if (res <= 0 || recv_subscribed(priv, *msg, ds_id) ||
-		    get_mgt_id(*msg) != ds_id) {
-			msg_put(*msg);
-			*msg = NULL;
-			continue;
-		}
-		priv->pmc_ds_requested = 0;
-		return 1;
-	}
-}
-
-static int run_pmc_wait_sync(struct phc2sys_private *priv, int timeout)
-{
-	struct ptp_message *msg;
-	int res;
-	void *data;
-	Enumeration8 portState;
-
-	while (1) {
-		res = run_pmc(priv, timeout, TLV_PORT_DATA_SET, &msg);
-		if (res <= 0)
-			return res;
-
-		data = get_mgt_data(msg);
-		portState = ((struct portDS *)data)->portState;
-		msg_put(msg);
-
-		switch (portState) {
-		case PS_MASTER:
-		case PS_SLAVE:
-			return 1;
-		}
-		/* try to get more data sets (for other ports) */
-		priv->pmc_ds_requested = 1;
-	}
-}
-
-static int run_pmc_get_utc_offset(struct phc2sys_private *priv, int timeout)
-{
-	struct ptp_message *msg;
-	int res;
-	struct timePropertiesDS *tds;
-
-	res = run_pmc(priv, timeout, TLV_TIME_PROPERTIES_DATA_SET, &msg);
-	if (res <= 0)
-		return res;
-
-	tds = (struct timePropertiesDS *)get_mgt_data(msg);
-	if (tds->flags & PTP_TIMESCALE) {
-		priv->sync_offset = tds->currentUtcOffset;
-		if (tds->flags & LEAP_61)
-			priv->leap = 1;
-		else if (tds->flags & LEAP_59)
-			priv->leap = -1;
-		else
-			priv->leap = 0;
-		priv->utc_offset_traceable = tds->flags & UTC_OFF_VALID &&
-					     tds->flags & TIME_TRACEABLE;
-	} else {
-		priv->sync_offset = 0;
-		priv->leap = 0;
-		priv->utc_offset_traceable = 0;
-	}
-	msg_put(msg);
-	return 1;
-}
-
-static int run_pmc_get_number_ports(struct phc2sys_private *priv, int timeout)
-{
-	struct ptp_message *msg;
-	int res;
-	struct defaultDS *dds;
-
-	res = run_pmc(priv, timeout, TLV_DEFAULT_DATA_SET, &msg);
-	if (res <= 0)
-		return res;
-
-	dds = (struct defaultDS *)get_mgt_data(msg);
-	res = dds->numberPorts;
-	msg_put(msg);
-	return res;
-}
-
-static int run_pmc_subscribe(struct phc2sys_private *priv, int timeout)
-{
-	struct ptp_message *msg;
-	int res;
-
-	res = run_pmc(priv, timeout, TLV_SUBSCRIBE_EVENTS_NP, &msg);
-	if (res <= 0)
-		return res;
-	msg_put(msg);
-	return 1;
-}
-
-static void run_pmc_events(struct phc2sys_private *priv)
-{
-	struct ptp_message *msg;
-
-	run_pmc(priv, 0, -1, &msg);
-}
-
-static int run_pmc_port_properties(struct phc2sys_private *priv, int timeout,
-				   unsigned int port,
-				   int *state, int *tstamping, char *iface)
-{
-	struct ptp_message *msg;
-	int res, len;
-	struct port_properties_np *ppn;
-
-	pmc_target_port(priv->pmc, port);
-	while (1) {
-		res = run_pmc(priv, timeout, TLV_PORT_PROPERTIES_NP, &msg);
-		if (res <= 0)
-			goto out;
-
-		ppn = get_mgt_data(msg);
-		if (ppn->portIdentity.portNumber != port) {
-			msg_put(msg);
-			continue;
-		}
-
-		*state = ppn->port_state;
-		*tstamping = ppn->timestamping;
-		len = ppn->interface.length;
-		if (len > IFNAMSIZ - 1)
-			len = IFNAMSIZ - 1;
-		memcpy(iface, ppn->interface.text, len);
-		iface[len] = '\0';
-
-		msg_put(msg);
-		res = 1;
-		break;
-	}
-out:
-	pmc_target_all(priv->pmc);
-	return res;
-}
-
-static int run_pmc_clock_identity(struct phc2sys_private *priv, int timeout)
-{
-	struct ptp_message *msg;
-	struct defaultDS *dds;
-	int res;
-
-	res = run_pmc(priv, timeout, TLV_DEFAULT_DATA_SET, &msg);
-	if (res <= 0)
-		return res;
-
-	dds = (struct defaultDS *)get_mgt_data(msg);
-	memcpy(&priv->clock_identity, &dds->clockIdentity,
-	       sizeof(struct ClockIdentity));
-	priv->clock_identity_set = 1;
-	msg_put(msg);
-	return 1;
-}
-
-static void close_pmc(struct phc2sys_private *priv)
-{
-	pmc_destroy(priv->pmc);
-	priv->pmc = NULL;
-}
-
 static int auto_init_ports(struct phc2sys_private *priv, int add_rt)
 {
-	struct port *port;
-	struct clock *clock;
-	int number_ports, res;
-	unsigned int i;
-	int state, timestamping;
+	int err, number_ports, phc_index, timestamping;
+	enum port_state state;
 	char iface[IFNAMSIZ];
+	struct clock *clock;
+	struct port *port;
+	unsigned int i;
 
 	while (1) {
-		if (!is_running())
+		if (!is_running()) {
 			return -1;
-		res = run_pmc_clock_identity(priv, 1000);
-		if (res < 0)
-			return -1;
-		if (res > 0)
+		}
+		err = pmc_agent_query_dds(priv->agent, 1000);
+		if (!err) {
 			break;
-		/* res == 0, timeout */
-		pr_notice("Waiting for ptp4l...");
+		}
+		if (err == -ETIMEDOUT) {
+			pr_notice("Waiting for ptp4l...");
+		} else {
+			return -1;
+		}
 	}
 
-	number_ports = run_pmc_get_number_ports(priv, 1000);
+	number_ports = pmc_agent_get_number_ports(priv->agent);
 	if (number_ports <= 0) {
 		pr_err("failed to get number of ports");
 		return -1;
 	}
 
-	res = run_pmc_subscribe(priv, 1000);
-	if (res <= 0) {
+	err = pmc_agent_subscribe(priv->agent, 1000);
+	if (err) {
 		pr_err("failed to subscribe");
 		return -1;
 	}
 
 	for (i = 1; i <= number_ports; i++) {
-		res = run_pmc_port_properties(priv, 1000, i, &state,
-					      &timestamping, iface);
-		if (res == -1) {
+		err = pmc_agent_query_port_properties(priv->agent, 1000, i,
+						      &state, &timestamping,
+						      &phc_index, iface);
+		if (err == -ENODEV) {
 			/* port does not exist, ignore the port */
 			continue;
 		}
-		if (res <= 0) {
+		if (err) {
 			pr_err("failed to get port properties");
 			return -1;
 		}
@@ -1206,10 +887,10 @@ static int auto_init_ports(struct phc2sys_private *priv, int add_rt)
 			/* ignore ports with software time stamping */
 			continue;
 		}
-		port = port_add(priv, i, iface);
+		port = port_add(priv, i, iface, phc_index);
 		if (!port)
 			return -1;
-		port->state = normalize_state(state);
+		port->state = port_state_normalize(state);
 	}
 	if (LIST_EMPTY(&priv->clocks)) {
 		pr_err("no suitable ports available");
@@ -1221,7 +902,7 @@ static int auto_init_ports(struct phc2sys_private *priv, int add_rt)
 	priv->state_changed = 1;
 
 	if (add_rt) {
-		clock = clock_add(priv, "CLOCK_REALTIME");
+		clock = clock_add(priv, "CLOCK_REALTIME", -1);
 		if (!clock)
 			return -1;
 		if (add_rt == 1)
@@ -1229,44 +910,20 @@ static int auto_init_ports(struct phc2sys_private *priv, int add_rt)
 	}
 
 	/* get initial offset */
-	if (run_pmc_get_utc_offset(priv, 1000) <= 0) {
+	if (pmc_agent_query_utc_offset(priv->agent, 1000)) {
 		pr_err("failed to get UTC offset");
 		return -1;
 	}
 	return 0;
 }
 
-/* Returns: -1 in case of error, 0 otherwise */
-static int update_pmc(struct phc2sys_private *priv, int subscribe)
-{
-	struct timespec tp;
-	uint64_t ts;
-
-	if (clock_gettime(CLOCK_MONOTONIC, &tp)) {
-		pr_err("failed to read clock: %m");
-		return -1;
-	}
-	ts = tp.tv_sec * NS_PER_SEC + tp.tv_nsec;
-
-	if (priv->pmc &&
-	    !(ts > priv->pmc_last_update &&
-	      ts - priv->pmc_last_update < PMC_UPDATE_INTERVAL)) {
-		if (subscribe)
-			run_pmc_subscribe(priv, 0);
-		if (run_pmc_get_utc_offset(priv, 0) > 0)
-			priv->pmc_last_update = ts;
-	}
-
-	return 0;
-}
-
 /* Returns: non-zero to skip clock update */
 static int clock_handle_leap(struct phc2sys_private *priv, struct clock *clock,
 			     int64_t offset, uint64_t ts)
 {
-	int clock_leap, node_leap = priv->leap;
+	int clock_leap, node_leap = pmc_agent_get_leap(priv->agent);
 
-	clock->sync_offset = priv->sync_offset;
+	clock->sync_offset = pmc_agent_get_sync_offset(priv->agent);
 
 	if ((node_leap || clock->leap_set) &&
 	    clock->is_utc != priv->master->is_utc) {
@@ -1307,7 +964,7 @@ static int clock_handle_leap(struct phc2sys_private *priv, struct clock *clock,
 		}
 	}
 
-	if (priv->utc_offset_traceable &&
+	if (pmc_agent_utc_offset_traceable(priv->agent) &&
 	    clock->utc_offset_set != clock->sync_offset) {
 		if (clock->clkid == CLOCK_REALTIME)
 			sysclk_set_tai_offset(clock->sync_offset);
@@ -1317,6 +974,55 @@ static int clock_handle_leap(struct phc2sys_private *priv, struct clock *clock,
 	return 0;
 }
 
+static bool hardpps_configured(int fd)
+{
+	return fd >= 0;
+}
+
+static int phc2sys_static_src_configuration(struct phc2sys_private *priv,
+					    const char *src_name)
+{
+	struct clock *src;
+
+	src = clock_add(priv, src_name, -1);
+	if (!src) {
+		fprintf(stderr, "valid source clock must be selected.\n");
+		return -1;
+	}
+	src->state = PS_SLAVE;
+	priv->master = src;
+
+	return 0;
+}
+
+static int phc2sys_static_dst_configuration(struct phc2sys_private *priv,
+					    const char *dst_name)
+{
+	struct clock *dst;
+
+	dst = clock_add(priv, dst_name, -1);
+	if (!dst) {
+		fprintf(stderr, "valid destination clock must be selected.\n");
+		return -1;
+	}
+	dst->state = PS_MASTER;
+	LIST_INSERT_HEAD(&priv->dst_clocks, dst, dst_list);
+
+	return 0;
+}
+
+static bool phc2sys_using_systemclock(struct phc2sys_private *priv)
+{
+	struct clock *c;
+
+	LIST_FOREACH(c, &priv->clocks, list) {
+		if (c->clkid == CLOCK_REALTIME) {
+			return true;
+		}
+	}
+	return false;
+}
+
 static void usage(char *progname)
 {
 	fprintf(stderr,
@@ -1328,10 +1034,10 @@ static void usage(char *progname)
 		" -r             synchronize system (realtime) clock\n"
 		"                repeat -r to consider it also as a time source\n"
 		" manual configuration:\n"
-		" -c [dev|name]  slave clock (CLOCK_REALTIME)\n"
-		" -d [dev]       master PPS device\n"
-		" -s [dev|name]  master clock\n"
-		" -O [offset]    slave-master time offset (0)\n"
+		" -c [dev|name]  time sink device (CLOCK_REALTIME)\n"
+		" -d [dev]       time source PPS device\n"
+		" -s [dev|name]  time source device\n"
+		" -O [offset]    sink-source time offset in seconds (0)\n"
 		" -w             wait for ptp4l\n"
 		" common options:\n"
 		" -f [file]      configuration file\n"
@@ -1340,8 +1046,8 @@ static void usage(char *progname)
 		" -I [ki]        integration constant (0.3)\n"
 		" -S [step]      step threshold (disabled)\n"
 		" -F [step]      step threshold only on start (0.00002)\n"
-		" -R [rate]      slave clock update rate in HZ (1.0)\n"
-		" -N [num]       number of master clock readings per update (5)\n"
+		" -R [rate]      update rate for the time sink devices in HZ (1.0)\n"
+		" -N [num]       number of source clock readings per update (5)\n"
 		" -L [limit]     sanity frequency limit in ppb (200000000)\n"
 		" -M [num]       NTP SHM segment number (0)\n"
 		" -u [num]       number of clock updates in summary stats (0)\n"
@@ -1360,12 +1066,14 @@ static void usage(char *progname)
 
 int main(int argc, char *argv[])
 {
-	char *config = NULL, *dst_name = NULL, *progname, *src_name = NULL;
-	struct clock *src, *dst;
+	char *config = NULL, *progname, *src_name = NULL;
+	const char *dst_names[MAX_DST_CLOCKS];
+	char uds_local[MAX_IFNAME_SIZE + 1];
+	int i, autocfg = 0, c, domain_number = 0, index, ntpshm_segment, offset;
+	int pps_fd = -1, print_level = LOG_INFO, r = -1, rt = 0;
+	int wait_sync = 0, dst_cnt = 0;
 	struct config *cfg;
 	struct option *opts;
-	int autocfg = 0, c, domain_number = 0, index, ntpshm_segment;
-	int pps_fd = -1, print_level = LOG_INFO, r = -1, rt = 0, wait_sync = 0;
 	double phc_rate, tmp;
 	struct phc2sys_private priv = {
 		.phc_readings = 5,
@@ -1378,6 +1086,10 @@ int main(int argc, char *argv[])
 	if (!cfg) {
 		return -1;
 	}
+	priv.agent = pmc_agent_create();
+	if (!priv.agent) {
+		return -1;
+	}
 
 	opts = config_long_options(cfg);
 
@@ -1403,7 +1115,11 @@ int main(int argc, char *argv[])
 			rt++;
 			break;
 		case 'c':
-			dst_name = strdup(optarg);
+			if (dst_cnt == MAX_DST_CLOCKS) {
+				fprintf(stderr, "too many sink clocks\n");
+				goto bad_usage;
+			}
+			dst_names[dst_cnt++] = optarg;
 			break;
 		case 'd':
 			pps_fd = open(optarg, O_RDONLY);
@@ -1421,7 +1137,7 @@ int main(int argc, char *argv[])
 				"'-i' has been deprecated. please use '-s' instead.\n");
             /* fallthrough */
 		case 's':
-			src_name = strdup(optarg);
+			src_name = optarg;
 			break;
 		case 'E':
 			if (!strcasecmp(optarg, "pi")) {
@@ -1433,6 +1149,9 @@ int main(int argc, char *argv[])
 			} else if (!strcasecmp(optarg, "ntpshm")) {
 				config_set_int(cfg, "clock_servo",
 					       CLOCK_SERVO_NTPSHM);
+			} else if (!strcasecmp(optarg, "refclock_sock")) {
+				config_set_int(cfg, "clock_servo",
+					       CLOCK_SERVO_REFCLOCK_SOCK);
 			} else {
 				fprintf(stderr,
 					"invalid servo name %s\n", optarg);
@@ -1469,9 +1188,10 @@ int main(int argc, char *argv[])
 				goto end;
 			break;
 		case 'O':
-			if (get_arg_val_i(c, optarg, &priv.sync_offset,
-					  INT_MIN, INT_MAX))
+			if (get_arg_val_i(c, optarg, &offset, INT_MIN, INT_MAX)) {
 				goto end;
+			}
+			pmc_agent_set_sync_offset(priv.agent, offset);
 			priv.forced_sync_offset = -1;
 			break;
 		case 'L':
@@ -1553,12 +1273,17 @@ int main(int argc, char *argv[])
 		return c;
 	}
 
-	if (autocfg && (src_name || dst_name || pps_fd >= 0 || wait_sync || priv.forced_sync_offset)) {
+	if (!autocfg && dst_cnt == 0) {
+		dst_names[dst_cnt++] = "CLOCK_REALTIME";
+	}
+
+	if (autocfg && (src_name || dst_cnt > 0 || hardpps_configured(pps_fd) ||
+			wait_sync || priv.forced_sync_offset)) {
 		fprintf(stderr,
 			"autoconfiguration cannot be mixed with manual config options.\n");
 		goto bad_usage;
 	}
-	if (!autocfg && pps_fd < 0 && !src_name) {
+	if (!autocfg && !hardpps_configured(pps_fd) && !src_name) {
 		fprintf(stderr,
 			"autoconfiguration or valid source clock must be selected.\n");
 		goto bad_usage;
@@ -1570,9 +1295,11 @@ int main(int argc, char *argv[])
 		goto bad_usage;
 	}
 
-	if (priv.servo_type == CLOCK_SERVO_NTPSHM) {
-		priv.kernel_leap = 0;
-		priv.sanity_freq_limit = 0;
+	if (hardpps_configured(pps_fd) && (dst_cnt != 1 ||
+	    strcmp(dst_names[0], "CLOCK_REALTIME"))) {
+		fprintf(stderr,
+			"cannot use a pps device unless destination is CLOCK_REALTIME\n");
+		goto bad_usage;
 	}
 
 	print_set_progname(progname);
@@ -1581,57 +1308,49 @@ int main(int argc, char *argv[])
 	print_set_syslog(config_get_int(cfg, NULL, "use_syslog"));
 	print_set_level(config_get_int(cfg, NULL, "logging_level"));
 
+	priv.free_running = config_get_int(cfg, NULL, "free_running");
 	priv.servo_type = config_get_int(cfg, NULL, "clock_servo");
-	if (priv.servo_type == CLOCK_SERVO_NTPSHM) {
+	if (priv.free_running || priv.servo_type == CLOCK_SERVO_NTPSHM) {
 		config_set_int(cfg, "kernel_leap", 0);
 		config_set_int(cfg, "sanity_freq_limit", 0);
 	}
 	priv.kernel_leap = config_get_int(cfg, NULL, "kernel_leap");
 	priv.sanity_freq_limit = config_get_int(cfg, NULL, "sanity_freq_limit");
 
+	for (i = 0; i < dst_cnt; i++) {
+		r = phc2sys_static_dst_configuration(&priv, dst_names[i]);
+		if (r) {
+			goto end;
+		}
+	}
+
+	snprintf(uds_local, sizeof(uds_local), "/var/run/phc2sys.%d",
+		 getpid());
+
 	if (autocfg) {
-		if (init_pmc(cfg, &priv))
+		if (init_pmc_node(cfg, priv.agent, uds_local,
+				  phc2sys_recv_subscribed, &priv))
 			goto end;
 		if (auto_init_ports(&priv, rt) < 0)
 			goto end;
-		r = do_loop(&priv, 1);
+		r = do_loop(&priv);
 		goto end;
 	}
 
-	src = clock_add(&priv, src_name);
-	free(src_name);
-	if (!src) {
-		fprintf(stderr,
-			"valid source clock must be selected.\n");
-		goto bad_usage;
-	}
-	src->state = PS_SLAVE;
-	priv.master = src;
-
-	dst = clock_add(&priv, dst_name ? dst_name : "CLOCK_REALTIME");
-	free(dst_name);
-	if (!dst) {
-		fprintf(stderr,
-			"valid destination clock must be selected.\n");
-		goto bad_usage;
-	}
-	dst->state = PS_MASTER;
-	LIST_INSERT_HEAD(&priv.dst_clocks, dst, dst_list);
-
-	if (pps_fd >= 0 && dst->clkid != CLOCK_REALTIME) {
-		fprintf(stderr,
-			"cannot use a pps device unless destination is CLOCK_REALTIME\n");
-		goto bad_usage;
+	r = phc2sys_static_src_configuration(&priv, src_name);
+	if (r) {
+		goto end;
 	}
 
 	r = -1;
 
 	if (wait_sync) {
-		if (init_pmc(cfg, &priv))
+		if (init_pmc_node(cfg, priv.agent, uds_local,
+				  phc2sys_recv_subscribed, &priv))
 			goto end;
 
 		while (is_running()) {
-			r = run_pmc_wait_sync(&priv, 1000);
+			r = run_pmc_wait_sync(priv.agent, 1000);
 			if (r < 0)
 				goto end;
 			if (r > 0)
@@ -1641,31 +1360,34 @@ int main(int argc, char *argv[])
 		}
 
 		if (!priv.forced_sync_offset) {
-			r = run_pmc_get_utc_offset(&priv, 1000);
-			if (r <= 0) {
+			r = pmc_agent_query_utc_offset(priv.agent, 1000);
+			if (r) {
 				pr_err("failed to get UTC offset");
 				goto end;
 			}
 		}
 
 		if (priv.forced_sync_offset ||
-		    (src->clkid != CLOCK_REALTIME && dst->clkid != CLOCK_REALTIME) ||
-		    src->clkid == CLOCK_INVALID)
-			close_pmc(&priv);
+		    !phc2sys_using_systemclock(&priv) ||
+		    hardpps_configured(pps_fd)) {
+			pmc_agent_disable(priv.agent);
+		}
 	}
 
-	if (pps_fd >= 0) {
+	if (hardpps_configured(pps_fd)) {
+		struct clock *dst = LIST_FIRST(&priv.dst_clocks);
+
 		/* only one destination clock allowed with PPS until we
 		 * implement a mean to specify PTP port to PPS mapping */
+		dst->servo = servo_add(&priv, dst);
 		servo_sync_interval(dst->servo, 1.0);
 		r = do_pps_loop(&priv, dst, pps_fd);
 	} else {
-		r = do_loop(&priv, 0);
+		r = do_loop(&priv);
 	}
 
 end:
-	if (priv.pmc)
-		close_pmc(&priv);
+	pmc_agent_destroy(priv.agent);
 	clock_cleanup(&priv);
 	port_cleanup(&priv);
 	config_destroy(cfg);
