@@ -23,6 +23,7 @@
 #include <net/if.h>
 #include <netinet/in.h>
 #include <netpacket/packet.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -39,6 +40,7 @@
 #include "config.h"
 #include "contain.h"
 #include "ether.h"
+#include "missing.h"
 #include "print.h"
 #include "raw.h"
 #include "sk.h"
@@ -53,47 +55,94 @@ struct raw {
 	int vlan;
 };
 
-#define OP_AND  (BPF_ALU | BPF_AND | BPF_K)
-#define OP_JEQ  (BPF_JMP | BPF_JEQ | BPF_K)
-#define OP_JUN  (BPF_JMP | BPF_JA)
-#define OP_LDB  (BPF_LD  | BPF_B   | BPF_ABS)
-#define OP_LDH  (BPF_LD  | BPF_H   | BPF_ABS)
-#define OP_RETK (BPF_RET | BPF_K)
-
 #define PTP_GEN_BIT 0x08 /* indicates general message, if set in message type */
 
-#define N_RAW_FILTER    12
-#define RAW_FILTER_TEST 9
+#define PRP_MIN_PACKET_LEN 70
+#define PRP_TRAILER_LEN 6
 
-static struct sock_filter raw_filter[N_RAW_FILTER] = {
-	{OP_LDH,  0, 0, OFF_ETYPE   },
-	{OP_JEQ,  0, 4, ETH_P_8021Q          }, /*f goto non-vlan block*/
-	{OP_LDH,  0, 0, OFF_ETYPE + 4        },
-	{OP_JEQ,  0, 7, ETH_P_1588           }, /*f goto reject*/
-	{OP_LDB,  0, 0, ETH_HLEN + VLAN_HLEN },
-	{OP_JUN,  0, 0, 2                    }, /*goto test general bit*/
-	{OP_JEQ,  0, 4, ETH_P_1588  }, /*f goto reject*/
-	{OP_LDB,  0, 0, ETH_HLEN    },
-	{OP_AND,  0, 0, PTP_GEN_BIT }, /*test general bit*/
-	{OP_JEQ,  0, 1, 0           }, /*0,1=accept event; 1,0=accept general*/
-	{OP_RETK, 0, 0, 1500        }, /*accept*/
-	{OP_RETK, 0, 0, 0           }, /*reject*/
+/*
+ * tcpdump -d \
+ * '   (ether[12:2] == 0x8100 and ether[12 + 4 :2] == 0x88F7 and ether[14+4 :1] & 0x8 == 0x8) '\
+ * 'or (ether[12:2] == 0x88F7 and                                ether[14   :1] & 0x8 == 0x8) '
+ *
+ * (000) ldh      [12]
+ * (001) jeq      #0x8100          jt 2    jf 7
+ * (002) ldh      [16]
+ * (003) jeq      #0x88f7          jt 4    jf 12
+ * (004) ldb      [18]
+ * (005) and      #0x8
+ * (006) jeq      #0x8             jt 11   jf 12
+ * (007) jeq      #0x88f7          jt 8    jf 12
+ * (008) ldb      [14]
+ * (009) and      #0x8
+ * (010) jeq      #0x8             jt 11   jf 12
+ * (011) ret      #262144
+ * (012) ret      #0
+*/
+static struct sock_filter raw_filter_vlan_norm_general[] = {
+	{ 0x28, 0, 0, 0x0000000c },
+	{ 0x15, 0, 5, 0x00008100 },
+	{ 0x28, 0, 0, 0x00000010 },
+	{ 0x15, 0, 8, 0x000088f7 },
+	{ 0x30, 0, 0, 0x00000012 },
+	{ 0x54, 0, 0, 0x00000008 },
+	{ 0x15, 4, 5, 0x00000008 },
+	{ 0x15, 0, 4, 0x000088f7 },
+	{ 0x30, 0, 0, 0x0000000e },
+	{ 0x54, 0, 0, 0x00000008 },
+	{ 0x15, 0, 1, 0x00000008 },
+	{ 0x6, 0, 0, 0x00040000 },
+	{ 0x6, 0, 0, 0x00000000 },
+};
+
+/*
+ * tcpdump -d \
+ *  '   (ether[12:2] == 0x8100 and ether[12 + 4 :2] == 0x88F7 and ether[14+4 :1] & 0x8 != 0x8) '\
+ *  'or (ether[12:2] == 0x88F7 and                                ether[14   :1] & 0x8 != 0x8) '
+ *
+ * (000) ldh      [12]
+ * (001) jeq      #0x8100          jt 2    jf 7
+ * (002) ldh      [16]
+ * (003) jeq      #0x88f7          jt 4    jf 12
+ * (004) ldb      [18]
+ * (005) and      #0x8
+ * (006) jeq      #0x8             jt 12   jf 11
+ * (007) jeq      #0x88f7          jt 8    jf 12
+ * (008) ldb      [14]
+ * (009) and      #0x8
+ * (010) jeq      #0x8             jt 12   jf 11
+ * (011) ret      #262144
+ * (012) ret      #0
+ */
+static struct sock_filter raw_filter_vlan_norm_event[] = {
+	{ 0x28, 0, 0, 0x0000000c },
+	{ 0x15, 0, 5, 0x00008100 },
+	{ 0x28, 0, 0, 0x00000010 },
+	{ 0x15, 0, 8, 0x000088f7 },
+	{ 0x30, 0, 0, 0x00000012 },
+	{ 0x54, 0, 0, 0x00000008 },
+	{ 0x15, 5, 4, 0x00000008 },
+	{ 0x15, 0, 4, 0x000088f7 },
+	{ 0x30, 0, 0, 0x0000000e },
+	{ 0x54, 0, 0, 0x00000008 },
+	{ 0x15, 1, 0, 0x00000008 },
+	{ 0x6, 0, 0, 0x00040000 },
+	{ 0x6, 0, 0, 0x00000000 },
 };
 
 static int raw_configure(int fd, int event, int index,
 			 unsigned char *addr1, unsigned char *addr2, int enable)
 {
-	int err1, err2, filter_test, option;
+	int err1, err2, option;
 	struct packet_mreq mreq;
-	struct sock_fprog prg = { N_RAW_FILTER, raw_filter };
+	struct sock_fprog prg;
 
-	filter_test = RAW_FILTER_TEST;
 	if (event) {
-		raw_filter[filter_test].jt = 0;
-		raw_filter[filter_test].jf = 1;
+		prg.len = ARRAY_SIZE(raw_filter_vlan_norm_event);
+		prg.filter = raw_filter_vlan_norm_event;
 	} else {
-		raw_filter[filter_test].jt = 1;
-		raw_filter[filter_test].jf = 0;
+		prg.len = ARRAY_SIZE(raw_filter_vlan_norm_general);
+		prg.filter = raw_filter_vlan_norm_general;
 	}
 
 	if (setsockopt(fd, SOL_SOCKET, SO_ATTACH_FILTER, &prg, sizeof(prg))) {
@@ -206,6 +255,64 @@ static void addr_to_mac(void *mac, struct address *addr)
 	memcpy(mac, &addr->sll.sll_addr, MAC_LEN);
 }
 
+/* Determines if the packet has Parallel Redundancy Protocol (PRP) trailer. */
+static bool has_prp_trailer(unsigned char *ptr, int cnt, int eth_hlen)
+{
+	unsigned short suffix_id, lane_size_field, lsdu_size;
+	int ptp_msg_len, trailer_start, padding_len;
+	struct ptp_header *hdr;
+
+	/* try to parse like a PTP message to find out the message length */
+	if (cnt < sizeof(struct ptp_header))
+		return false;
+
+	hdr = (struct ptp_header *)ptr;
+	if ((hdr->ver & MAJOR_VERSION_MASK) != PTP_MAJOR_VERSION)
+		return false;
+
+	ptp_msg_len = ntohs(hdr->messageLength);
+
+	/* PRP requires ethernet packets to be minimum 70 bytes, including trailer */
+	trailer_start = ptp_msg_len;
+	padding_len = 0;
+	if ((eth_hlen + ptp_msg_len + PRP_TRAILER_LEN) < PRP_MIN_PACKET_LEN)
+	{
+		padding_len = PRP_MIN_PACKET_LEN - (eth_hlen + ptp_msg_len + PRP_TRAILER_LEN);
+		trailer_start += padding_len;
+	}
+
+	if (cnt < (trailer_start + PRP_TRAILER_LEN))
+		return false;
+
+	/* PRP trailer (RCT) consists of 3 uint16.
+	 | -------------------------------------------------------- |
+	 | SeqNr(0-15) | LanId(0-3) LSDUsize(4-15) | Suffix (0-15)  |
+	 | -------------------------------------------------------- |
+	 - Sequence number is a running number and can't be verified
+	 - LanId should be 0x1010 or 0x1011 (but should not be used for verification)
+	 - LSDUsize should match LSDU length
+	   (including possible padding and the RCT itself)
+	 - Suffix should be 0x88FB
+	*/
+
+	/* Verify that the size in the RCT matches.
+	   Size is the lower 12 bits
+	*/
+	lane_size_field = ntohs(*(unsigned short*)(ptr + trailer_start + 2));
+	lsdu_size = lane_size_field & 0x0FFF;
+	if (lsdu_size != cnt)
+		return false;
+
+	/* Verify the suffix */
+	suffix_id = ntohs(*(unsigned short*)(ptr + trailer_start + 4));
+	if (suffix_id == ETH_P_PRP)
+	{
+		return true;
+	}
+
+	return false;
+}
+
 static int raw_open(struct transport *t, struct interface *iface,
 		    struct fdarray *fda, enum timestamp_type ts_type)
 {
@@ -243,7 +350,8 @@ static int raw_open(struct transport *t, struct interface *iface,
 	if (gfd < 0)
 		goto no_general;
 
-	if (sk_timestamping_init(efd, name, ts_type, TRANS_IEEE_802_3))
+	if (sk_timestamping_init(efd, name, ts_type, TRANS_IEEE_802_3,
+				 interface_get_vclock(iface)))
 		goto no_timestamping;
 
 	if (sk_general_init(gfd))
@@ -265,10 +373,10 @@ no_mac:
 static int raw_recv(struct transport *t, int fd, void *buf, int buflen,
 		    struct address *addr, struct hw_timestamp *hwts)
 {
-	int cnt, hlen;
+	struct raw *raw = container_of(t, struct raw, t);
 	unsigned char *ptr = buf;
 	struct eth_hdr *hdr;
-	struct raw *raw = container_of(t, struct raw, t);
+	int cnt, hlen;
 
 	if (raw->vlan) {
 		hlen = sizeof(struct vlan_hdr);
@@ -286,6 +394,9 @@ static int raw_recv(struct transport *t, int fd, void *buf, int buflen,
 	if (cnt < 0)
 		return cnt;
 
+	if (has_prp_trailer(buf, cnt, hlen))
+		cnt -= PRP_TRAILER_LEN;
+
 	if (raw->vlan) {
 		if (ETH_P_1588 == ntohs(hdr->type)) {
 			pr_notice("raw: disabling VLAN mode");
