@@ -26,7 +26,9 @@
 #include "address.h"
 #include "ddt.h"
 #include "ether.h"
+#include "fsm.h"
 #include "transport.h"
+#include "unicast_fsm.h"
 
 #define MAX_PRINT_BYTES 16
 #define BIN_BUF_SIZE (MAX_PRINT_BYTES * 3 + 1)
@@ -57,13 +59,20 @@ const char *ts_str(enum timestamp_type ts);
  */
 int addreq(enum transport_type type, struct address *a, struct address *b);
 
-static inline uint16_t align16(uint16_t *p)
+static inline uint16_t align16(void *p)
 {
 	uint16_t v;
 	memcpy(&v, p, sizeof(v));
 	return v;
 }
 
+static inline uint32_t align32(void *p)
+{
+	uint32_t v;
+	memcpy(&v, p, sizeof(v));
+	return v;
+}
+
 char *bin2str_impl(Octet *data, int len, char *buf, int buf_len);
 
 /**
@@ -110,6 +119,15 @@ char *pid2str(struct PortIdentity *id);
 
 char *portaddr2str(struct PortAddress *addr);
 
+const char *ustate2str(enum unicast_state ustate);
+
+/**
+ * Reduce all port states for which the sync direction isn't known to
+ * PS_DISABLED, and report the given port state otherwise. This minimizes port
+ * state transitions for PMC agents when nothing interesting happened.
+ */
+enum port_state port_state_normalize(enum port_state state);
+
 /**
  * Closes a dynamic posix clock.
  * @param clock  A clock ID obtained via posix_clock_close().
